/* -LICENSE-START-
 ** Copyright (c) 2018 Blackmagic Design
 **
 ** Permission is hereby granted, free of charge, to any person or organization
 ** obtaining a copy of the software and accompanying documentation covered by
 ** this license (the "Software") to use, reproduce, display, distribute,
 ** execute, and transmit the Software, and to prepare derivative works of the
 ** Software, and to permit third-parties to whom the Software is furnished to
 ** do so, all subject to the following:
 **
 ** The copyright notices in the Software and this entire statement, including
 ** the above license grant, this restriction and the following disclaimer,
 ** must be included in all copies of the Software, in whole or in part, and
 ** all derivative works of the Software, unless such copies or derivative
 ** works are solely in the form of machine-executable object code generated by
 ** a source language processor.
 **
 ** THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 ** IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 ** FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. IN NO EVENT
 ** SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE
 ** FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR OTHERWISE,
 ** ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 ** DEALINGS IN THE SOFTWARE.
 ** -LICENSE-END-
 */

#include "BlackmagicRawAPI.h"

#include <cstdint>
#include <atomic>
#include <chrono>
#include <iostream>
#include <thread>

#include <CoreServices/CoreServices.h>
#include <ImageIO/ImageIO.h>

#ifdef DEBUG
	#include <cassert>
	#define VERIFY(condition) assert(SUCCEEDED(condition))
#else
	#define VERIFY(condition) condition
#endif

static const BlackmagicRawResourceFormat	s_resourceFormat	= blackmagicRawResourceFormatRGBAU8;
static const int							s_maxJobsInFlight	= 3;
static std::atomic<int>						s_jobsInFlight		= {0};
static IBlackmagicRawPipelineDevice*		s_openclDevice		= nullptr;
static IBlackmagicRawResourceManager 		*s_resourceManager	= nullptr;

struct UserData
{
	unsigned long long frameIndex;
};

static void OutputImage(uint32_t width, uint32_t height, uint32_t sizeBytes, void* imageData)
{
	static const size_t s_bitsPerComponent = 8;
	static const size_t s_bitsPerPixel = 32;
	static const CFStringRef s_outputFileName = CFSTR("/tmp/outputRGBAU8.png");

	size_t bytesPerRow = (s_bitsPerPixel * width) / 8;
	CGColorSpaceRef space = CGColorSpaceCreateWithName(kCGColorSpaceSRGB);
	CGBitmapInfo bitmapInfo = kCGImageAlphaNoneSkipLast | kCGImageByteOrderDefault;
	CGDataProviderRef provider = CGDataProviderCreateWithData(nullptr, imageData, sizeBytes, nullptr);
	const CGFloat* decode = nullptr;
	bool shouldInterpolate = false;
	CGColorRenderingIntent intent = kCGRenderingIntentDefault;

	CGImageRef imageRef = CGImageCreate(width, height, s_bitsPerComponent, s_bitsPerPixel, bytesPerRow, space, bitmapInfo, provider, decode, shouldInterpolate, intent);

	CFURLRef file = CFURLCreateWithFileSystemPath(kCFAllocatorDefault, s_outputFileName, kCFURLPOSIXPathStyle, false);
	if (file == nullptr)
	{
		std::cerr << "Failed to create " << CFStringGetCStringPtr(s_outputFileName, kCFStringEncodingMacRoman) << "!" << std::endl;
		return;
	}
	else
	{
		std::cerr << "Created " << CFStringGetCStringPtr(s_outputFileName, kCFStringEncodingMacRoman) << " (" << width << " x " << height << " pixels)!" << std::endl;
	}

	CGImageDestinationRef destination = CGImageDestinationCreateWithURL(file, kUTTypePNG, 1, nullptr);
	if (destination)
	{
		CGImageDestinationAddImage(destination, imageRef, nil);
		CGImageDestinationFinalize(destination);
	}

	CFRelease(destination);
	CFRelease(file);
	CGImageRelease(imageRef);
	CGDataProviderRelease(provider);
	CGColorSpaceRelease(space);
}

class CameraCodecCallback : public IBlackmagicRawCallback
{
public:
	explicit CameraCodecCallback() = default;
	virtual ~CameraCodecCallback() = default;

	virtual void ReadComplete(IBlackmagicRawJob* readJob, HRESULT result, IBlackmagicRawFrame* frame)
	{
		UserData* userData = nullptr;
		VERIFY(readJob->GetUserData((void**)&userData));

		IBlackmagicRawJob* decodeAndProcessJob = nullptr;

		if (result == S_OK)
			VERIFY(frame->SetResourceFormat(s_resourceFormat));

		if (result == S_OK)
			result = frame->CreateJobDecodeAndProcessFrame(nullptr, nullptr, &decodeAndProcessJob);

		if (result == S_OK)
			VERIFY(decodeAndProcessJob->SetUserData(userData));

		if (result == S_OK)
			result = decodeAndProcessJob->Submit();

		if (result != S_OK)
		{
			if (decodeAndProcessJob)
				decodeAndProcessJob->Release();

			delete userData;
		}

		readJob->Release();
	}

	virtual void ProcessComplete(IBlackmagicRawJob* job, HRESULT result, IBlackmagicRawProcessedImage* processedImage)
	{
		UserData* userData = nullptr;
		VERIFY(job->GetUserData((void**)&userData));

		std::cout << "Processed frame index: " << userData->frameIndex << std::endl;
		delete userData;

		job->Release();
		--s_jobsInFlight;

		
		uint32_t processedImageSizeBytes = 0;
		uint32_t processedImageWidthPixels = 0;
		uint32_t processedImageHeightPixels = 0;
		void* processedResourceOnGPU = nullptr;
		VERIFY(processedImage->GetWidth(&processedImageWidthPixels) == S_OK);
		VERIFY(processedImage->GetHeight(&processedImageHeightPixels) == S_OK);
		VERIFY(processedImage->GetResourceSizeBytes(&processedImageSizeBytes) == S_OK);
		VERIFY(processedImage->GetResource(&processedResourceOnGPU) == S_OK);
		
		void *cpuImage = nullptr;
		
		BlackmagicRawPipeline pipeline;
		void* context;
		void* commandQueue;
		VERIFY(s_openclDevice->GetPipeline(&pipeline, &context, &commandQueue) == S_OK);
		
		VERIFY(s_resourceManager->CreateResource(nullptr, commandQueue, processedImageSizeBytes, blackmagicRawResourceTypeBufferCPU, blackmagicRawResourceUsageReadCPUWriteCPU, &cpuImage) == S_OK);
		
		VERIFY(s_resourceManager->CopyResource(nullptr, commandQueue, processedResourceOnGPU, blackmagicRawResourceTypeBufferOpenCL, cpuImage, blackmagicRawResourceTypeBufferCPU, processedImageSizeBytes, false) == S_OK);
		
		// Save the image to disk using
		OutputImage(processedImageWidthPixels, processedImageHeightPixels, processedImageSizeBytes, cpuImage);
		
		s_resourceManager->ReleaseResource(nullptr, commandQueue, cpuImage, blackmagicRawResourceTypeBufferCPU);
	}

	virtual void DecodeComplete(IBlackmagicRawJob*, HRESULT) {}
	virtual void TrimProgress(IBlackmagicRawJob*, float) {}
	virtual void TrimComplete(IBlackmagicRawJob*, HRESULT) {}
	virtual void SidecarMetadataParseWarning(IBlackmagicRawClip*, CFStringRef, uint32_t, CFStringRef) {}
	virtual void SidecarMetadataParseError(IBlackmagicRawClip*, CFStringRef, uint32_t, CFStringRef) {}
	virtual void PreparePipelineComplete(void*, HRESULT) {}

	virtual HRESULT STDMETHODCALLTYPE QueryInterface(REFIID, LPVOID*)
	{
		return E_NOTIMPL;
	}

	virtual ULONG STDMETHODCALLTYPE AddRef(void)
	{
		return 0;
	}

	virtual ULONG STDMETHODCALLTYPE Release(void)
	{
		return 0;
	}
};

HRESULT ProcessClip(IBlackmagicRawClip* clip)
{
	HRESULT result;

	unsigned long long frameCount = 0;
	unsigned long long frameIndex = 0;

	result = clip->GetFrameCount(&frameCount);

	while (frameIndex < frameCount)
	{
		if (s_jobsInFlight >= s_maxJobsInFlight)
		{
			std::this_thread::sleep_for(std::chrono::microseconds(100));
			continue;
		}

		IBlackmagicRawJob* jobRead = nullptr;
		if (result == S_OK)
			result = clip->CreateJobReadFrame(frameIndex, &jobRead);

		UserData* userData = nullptr;
		if (result == S_OK)
		{
			userData = new UserData();
			userData->frameIndex = frameIndex;
			VERIFY(jobRead->SetUserData(userData));
		}

		if (result == S_OK)
			result = jobRead->Submit();

		if (result != S_OK)
		{
			if (userData != nullptr)
				delete userData;

			if (jobRead != nullptr)
				jobRead->Release();

			break;
		}

		++s_jobsInFlight;

		frameIndex++;
	}

	return result;
}

int main(int argc, const char* argv[])
{
	if (argc > 2)
	{
		std::cerr << "Usage: " << argv[0] << " clipName.braw" << std::endl;
		return 1;
	}

	CFStringRef clipName;
	bool clipNameProvided = argc == 2;
	if (clipNameProvided)
	{
		clipName = CFStringCreateWithCString(NULL, argv[1], kCFStringEncodingUTF8);
	}
	else
	{
		clipName = CFSTR("../../../Media/sample.braw");
	}

	HRESULT result = S_OK;

	IBlackmagicRawFactory* factory = nullptr;
	IBlackmagicRaw* codec = nullptr;
	IBlackmagicRawClip* clip = nullptr;
	IBlackmagicRawConfiguration* config = nullptr;
	IBlackmagicRawConfigurationEx *configEx = nullptr;

	CameraCodecCallback callback;

	do
	{
		factory = CreateBlackmagicRawFactoryInstanceFromPath(CFSTR("../../Libraries"));
		if (factory == nullptr)
		{
			std::cerr << "Failed to create IBlackmagicRawFactory!" << std::endl;
			break;
		}

		result = factory->CreateCodec(&codec);
		if (result != S_OK)
		{
			std::cerr << "Failed to create IBlackmagicRaw!" << std::endl;
			break;
		}

		IBlackmagicRawPipelineDeviceIterator* pipelineDeviceIterator = nullptr;
		result = factory->CreatePipelineDeviceIterator(blackmagicRawPipelineOpenCL, blackmagicRawInteropNone, &pipelineDeviceIterator);
		if (result != S_OK)
		{
			std::cerr << "Failed to create device iterator for blackmagicRawPipelineOpenCL with blackmagicRawInteropNone!" << std::endl;
			break;
		}

		// Create the (first) device
		result = pipelineDeviceIterator->CreateDevice(&s_openclDevice);
		if (result != S_OK)
		{
			std::cerr << "Failed to create the first OpenCL device as indicated by pipelineDeviceIterator!" << std::endl;
			break;
		}

		// Release the device iterator as it is no longer required
		pipelineDeviceIterator->Release();

		result = codec->QueryInterface(IID_IBlackmagicRawConfiguration, (void**)&config);
		if (result != S_OK)
		{
			std::cerr << "Failed to get IID_IBlackmagicRawConfiguration!" << std::endl;
			break;
		}

		result = config->SetFromDevice(s_openclDevice);
		if (result != S_OK)
		{
			std::cerr << "Failed to set the config from the OpenCL device!" << std::endl;
			break;
		}

		result = codec->QueryInterface(IID_IBlackmagicRawConfigurationEx, (void **)&configEx);
		if (result != S_OK)
		{
			std::cerr << "Failed to get IID_IBlackmagicRawConfigurationEx!" << std::endl;
			break;
		}

		result = configEx->GetResourceManager(&s_resourceManager);
		if (result != S_OK)
		{
			std::cerr << "Failed to get Resource Manager!" << std::endl;
			break;
		}

		result = codec->OpenClip(clipName, &clip);
		if (result != S_OK)
		{
			std::cerr << "Failed to open IBlackmagicRawClip!" << std::endl;
			break;
		}

		result = codec->SetCallback(&callback);
		if (result != S_OK)
		{
			std::cerr << "Failed to set IBlackmagicRawCallback!" << std::endl;
			break;
		}

		result = ProcessClip(clip);
		codec->FlushJobs();

	} while(0);

	if (config != nullptr)
		config->Release();

	if (clip != nullptr)
		clip->Release();

	if (codec != nullptr)
		codec->Release();

	s_openclDevice->Release();
	s_openclDevice = nullptr;

	if (factory != nullptr)
		factory->Release();

	CFRelease(clipName);

	return result;
}
